{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "[Day 23](https://adventofcode.com/2019/day/23) puzzle input is an IntCode program (mine is [here](puzzle_input/day23.txt)) which represents a networked computer as part of a bigger network. Part 1 involves connecting a number of these network computers, running them and finding the first packet sent to computer 255. Part 2 involves modifying the behaviour so that computer 255 does something when the network is \"idle\" (no packets are sent) which kicks things into action again, and finding the first `y` value sent to this computer twice in a row."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 2,
   "metadata": {},
   "outputs": [],
   "source": [
    "opcode_add = 1\n",
    "opcode_mul = 2\n",
    "opcode_read = 3\n",
    "opcode_write = 4\n",
    "opcode_jump_true = 5\n",
    "opcode_jump_false = 6\n",
    "opcode_lt = 7\n",
    "opcode_eq = 8\n",
    "opcode_rebase = 9\n",
    "\n",
    "opcode_terminate = 99\n",
    "\n",
    "mode_position = 0\n",
    "mode_immediate = 1\n",
    "mode_relative = 2\n",
    "\n",
    "class IntCodeCpu:\n",
    "    def __init__(self, memory_image):\n",
    "        self.memory = [ x for x in memory_image ] # copy memory image, in case it's reused\n",
    "        self.stalled = True\n",
    "        self.input_buffer = None\n",
    "        self.output_buffer = None\n",
    "        self.pc = 0\n",
    "        self.initialise_opcodes()\n",
    "        self.offset = 0\n",
    "        self.done = False\n",
    "        \n",
    "    def start(self, input_buffer, output_buffer, noun=None, verb=None):\n",
    "        self.input_buffer = input_buffer\n",
    "        self.output_buffer = output_buffer\n",
    "\n",
    "        if noun:\n",
    "            self.memory[1] = noun\n",
    "        if verb:\n",
    "            self.memory[2] = verb\n",
    "           \n",
    "        return self.run()\n",
    "            \n",
    "            \n",
    "    def run(self):\n",
    "        instr = self.memory[self.pc]\n",
    "        self.stalled = False\n",
    "\n",
    "        while int(instr) != opcode_terminate and not self.stalled:\n",
    "            (op, modes) = self.decode_instr(instr)\n",
    "            self.pc = op(modes)\n",
    "            instr = self.memory[self.pc]\n",
    "\n",
    "        return self.memory[0]\n",
    "\n",
    "    #-HELPERS-----------------------------\n",
    "    def try_pop_mode(self, modes):\n",
    "        if len(modes) == 0:\n",
    "            return 0\n",
    "    \n",
    "        return modes.pop()\n",
    "    \n",
    "    def resize_memory(self, target_addr):\n",
    "        self.memory += ([0] * (1 + target_addr - len(self.memory)))\n",
    "    \n",
    "    #-DECODE-INSTRUCTIONS-----------------\n",
    "    def initialise_opcodes(self):\n",
    "        self.opcodes = {\n",
    "            opcode_add: self.op_add,\n",
    "            opcode_mul: self.op_mul,\n",
    "            opcode_read: self.op_read,\n",
    "            opcode_write: self.op_write,\n",
    "            opcode_jump_true: self.op_jump_true,\n",
    "            opcode_jump_false: self.op_jump_false,\n",
    "            opcode_lt: self.op_lt,\n",
    "            opcode_eq: self.op_eq,\n",
    "            opcode_rebase: self.op_rebase\n",
    "        }    \n",
    "    \n",
    "    def decode_instr(self, instr):\n",
    "        instr = str(instr)\n",
    "        opcode = self.decode_op(instr)\n",
    "        modes = self.decode_modes(instr)\n",
    "\n",
    "        if not (opcode in self.opcodes):\n",
    "            raise Exception(f\"Invalid opcode {opcode}\")\n",
    "\n",
    "        return (self.opcodes[opcode], modes)    \n",
    "    \n",
    "    def decode_op(self, instr):\n",
    "        if len(instr) > 2:\n",
    "            return int(instr[-2:])\n",
    "        return int(instr)\n",
    "    \n",
    "    def decode_modes(self, instr):\n",
    "        if len(instr) > 2:\n",
    "            return [ int(d) for d in instr[:-2]]\n",
    "        return []\n",
    "    \n",
    "    #-MICRO-OPS---------------------------\n",
    "    def uop_read(self, value, mode):\n",
    "        if mode == mode_position:\n",
    "            if value >= len(self.memory):\n",
    "                self.resize_memory(value)        \n",
    "\n",
    "            return int(self.memory[value])\n",
    "\n",
    "        elif mode == mode_relative:\n",
    "            if self.offset + value >= len(self.memory):\n",
    "                self.resize_memory(self.offset + value)     \n",
    "\n",
    "            return int(self.memory[self.offset + value])\n",
    "\n",
    "        elif mode == mode_immediate:\n",
    "            return int(value)\n",
    "\n",
    "        else:\n",
    "            raise Exception(\"UNKNOWN MODE\")\n",
    "\n",
    "    def uop_write(self, dst, value, mode):\n",
    "        if mode == mode_position:\n",
    "            if dst >= len(self.memory):\n",
    "                self.resize_memory(dst)\n",
    "            self.memory[dst] = value\n",
    "\n",
    "        elif mode == mode_relative:\n",
    "            if self.offset + dst >= len(self.memory):\n",
    "                self.resize_memory(self.offset + dst)     \n",
    "\n",
    "            self.memory[self.offset + dst] = value\n",
    "\n",
    "        elif mode == mode_immediate:\n",
    "            raise Exception(f\"cannot write {value} to literal {dst}\")\n",
    "\n",
    "\n",
    "    def uop_cond_jump(self, modes, cond):\n",
    "        param_mode = self.try_pop_mode(modes)\n",
    "        param_raw = int(self.memory[self.pc + 1])\n",
    "        param = self.uop_read(param_raw, param_mode)\n",
    "\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest_raw = int(self.memory[self.pc + 2])\n",
    "        dest = self.uop_read(dest_raw, dest_mode)\n",
    "\n",
    "        if cond(param):\n",
    "            return dest\n",
    "\n",
    "        return self.pc + 3\n",
    "\n",
    "    def uop_cmp(self, modes, cmp):\n",
    "\n",
    "        param0_mode = self.try_pop_mode(modes)\n",
    "        param0_raw = int(self.memory[self.pc + 1])\n",
    "        param0 = self.uop_read(param0_raw, param0_mode)\n",
    "\n",
    "        param1_mode = self.try_pop_mode(modes)\n",
    "        param1_raw = int(self.memory[self.pc + 2])\n",
    "        param1 = self.uop_read(param1_raw, param1_mode)\n",
    "\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        if cmp(param0, param1):\n",
    "            self.uop_write(dest, 1, dest_mode)\n",
    "        else:\n",
    "            self.uop_write(dest, 0, dest_mode)\n",
    "\n",
    "        return self.pc + 4\n",
    "        \n",
    "    \n",
    "    #-OPCODES-----------------------------\n",
    "    def op_add(self, modes):\n",
    "        arg0_mode = self.try_pop_mode(modes)\n",
    "        arg1_mode = self.try_pop_mode(modes)\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "\n",
    "        arg0_raw = int(self.memory[self.pc + 1])\n",
    "        arg1_raw = int(self.memory[self.pc + 2])\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        arg0 = self.uop_read(arg0_raw, arg0_mode)\n",
    "        arg1 = self.uop_read(arg1_raw, arg1_mode)\n",
    "\n",
    "        self.uop_write(dest, str(int(arg0) + int(arg1)), dest_mode)\n",
    "        return self.pc + 4\n",
    "\n",
    "    def op_mul(self, modes):\n",
    "        arg0_mode = self.try_pop_mode(modes)\n",
    "        arg1_mode = self.try_pop_mode(modes)\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "\n",
    "        arg0_raw = int(self.memory[self.pc + 1])\n",
    "        arg1_raw = int(self.memory[self.pc + 2])\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        arg0 = self.uop_read(arg0_raw, arg0_mode)\n",
    "        arg1 = self.uop_read(arg1_raw, arg1_mode)\n",
    "\n",
    "        self.uop_write(dest, str(int(arg0) * int(arg1)), dest_mode)    \n",
    "        return self.pc + 4\n",
    "\n",
    "\n",
    "    def op_read(self, modes):\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest = int(self.memory[self.pc + 1])\n",
    "        \n",
    "        # if the input buffer is empty, we should \"stall\" and \n",
    "        # resume later\n",
    "        if not self.input_buffer:\n",
    "            self.stalled = True\n",
    "            return self.pc\n",
    "        \n",
    "        val = self.input_buffer.pop()\n",
    "\n",
    "        self.uop_write(dest, str(val), dest_mode)\n",
    "\n",
    "        return self.pc + 2\n",
    "\n",
    "    def op_write(self, modes):\n",
    "        src_mode = self.try_pop_mode(modes)\n",
    "        src_raw = int(self.memory[self.pc + 1])\n",
    "        src = self.uop_read(src_raw, src_mode)\n",
    "\n",
    "        self.output_buffer.append(src)\n",
    "\n",
    "        return self.pc + 2\n",
    "\n",
    "    def op_jump_true(self, modes):\n",
    "        return self.uop_cond_jump(modes, lambda x: x != 0)\n",
    "\n",
    "\n",
    "    def op_jump_false(self, modes):\n",
    "        return self.uop_cond_jump(modes, lambda x: x == 0)\n",
    "\n",
    "    def op_lt(self, modes):\n",
    "        return self.uop_cmp(modes, lambda x, y: x < y)\n",
    "\n",
    "    def op_eq(self, modes):\n",
    "        return self.uop_cmp(modes, lambda x, y: x == y)\n",
    "\n",
    "    def op_rebase(self, modes):    \n",
    "        param_mode = self.try_pop_mode(modes)\n",
    "        param_raw = int(self.memory[self.pc + 1])\n",
    "        param = self.uop_read(param_raw, param_mode)\n",
    "\n",
    "        self.offset += param\n",
    "\n",
    "        return self.pc + 2"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [],
   "source": [
    "from collections import defaultdict\n",
    "\n",
    "def network_send(net, address, packet):\n",
    "    if address == 255:\n",
    "        net[address] = [ packet ]\n",
    "    else:\n",
    "        net[address].append(packet)\n",
    "    \n",
    "def network_recv(net, address):\n",
    "    if len(net[address]) > 0:\n",
    "        return net[address].pop(0)\n",
    "    return [-1]"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 4,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "15969\n"
     ]
    }
   ],
   "source": [
    "raw_code = open(\"puzzle_input/day23.txt\", \"r\").read()\n",
    "code = raw_code.split(\",\")\n",
    "\n",
    "def part_one():\n",
    "    nodes = [ IntCodeCpu(code) for n in range(50) ]\n",
    "    network = defaultdict(list)\n",
    "\n",
    "    # init nodes\n",
    "    for addr, node in enumerate(nodes):\n",
    "        packet_in = [addr]\n",
    "        packet_out = []\n",
    "        node.start(packet_in, packet_out)\n",
    "\n",
    "        if len(packet_out) > 0:\n",
    "            dest, x, y = output\n",
    "            network_send(network, dest, [y, x])\n",
    "\n",
    "    while len(network[255]) == 0:\n",
    "        for addr, node in enumerate(nodes):\n",
    "            packet_in = network_recv(network, addr)\n",
    "            packet_out = []\n",
    "            node.start(packet_in, packet_out)\n",
    "\n",
    "            while len(packet_out) > 0:\n",
    "                dest, x, y = packet_out[:3]\n",
    "                network_send(network, dest, [y, x])\n",
    "                packet_out = packet_out[3:]\n",
    "                \n",
    "    return network[255][0][0]\n",
    "            \n",
    "print(part_one())"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 5,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "10650\n"
     ]
    }
   ],
   "source": [
    "def network_idle(net):\n",
    "    # determine if everyone's input buffer is idle (except 255 ofc)\n",
    "    return all([ \n",
    "        len(net[addr]) == 0 for addr in net.keys() if addr != 255 \n",
    "    ])    \n",
    "\n",
    "def part_two():\n",
    "    nodes = [ IntCodeCpu(code) for n in range(50) ]\n",
    "    network = defaultdict(list)\n",
    "\n",
    "    # init nodes\n",
    "    for addr, node in enumerate(nodes):\n",
    "        packet_in = [addr]\n",
    "        packet_out = []\n",
    "        node.start(packet_in, packet_out)\n",
    "\n",
    "        if len(packet_out) > 0:\n",
    "            dest, x, y = output\n",
    "            network_send(network, dest, [y, x])\n",
    "    \n",
    "    prev_y = None\n",
    "    while True:\n",
    "        for addr, node in enumerate(nodes):\n",
    "            packet_in = network_recv(network, addr)\n",
    "            packet_out = []\n",
    "            node.start(packet_in, packet_out)\n",
    "\n",
    "            while len(packet_out) > 0:\n",
    "                dest, x, y = packet_out[:3]\n",
    "                network_send(network, dest, [y, x])\n",
    "                packet_out = packet_out[3:]\n",
    "                \n",
    "        if network_idle(network):\n",
    "            y, x = network_recv(network, 255)\n",
    "            if prev_y == y:\n",
    "                return y\n",
    "            \n",
    "            network_send(network, 0, [y, x])\n",
    "            prev_y = y\n",
    "            \n",
    "print(part_two())"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 4
}
